/**
 * Copyright (c) 2006 Sascha Grau
 * contact: rostar@graui.de
 */

#include <iostream>

template<class Entry>
DLList<Entry>::DLList() {

    first = NULL;
    last = NULL;
    elements = 0;
}

template<class Entry>
DLList<Entry>::~DLList() {

    clear();
}

template<class Entry>
void DLList<Entry>::clear() {

    while (!empty())
        remove(last);

    first = NULL;
    last  = NULL;
}

/**
 * A helper function to hold shared code of both
 * push_back and push_front.
 * It adds the new entry between back and front.
 */
template<class Entry>
void DLList<Entry>::ring_insert(Entry e) {

    Element<Entry>* elem = new Element<Entry>(e);

    if (empty()) {
        first = elem;
        last  = elem;
        elem->next = elem;
        elem->prev = elem;
    } else {
        last->next = elem;
        elem->prev = last;
        first->prev = elem;
        elem->next = first;
    }

    elements++;
}

template<class Entry>
void DLList<Entry>::push_back(Entry e) {

    ring_insert(e);
    last = last->next;
}

template<class Entry>
void DLList<Entry>::push_front(Entry e) {

    ring_insert(e);
    first = first->prev;
}

template<class Entry>
Entry DLList<Entry>::at(unsigned long pos) const{

    if ((pos > elements-1) || (pos < 0) || (empty()))
        return NULL;

    int dec = 1;
    Element<Entry>* current = first;

    if (pos > elements / 2) {
        dec = -1;
        current = last;
        pos = pos - elements;
    }

    while (pos != 0) {
        pos = pos - dec;
        current = (dec > 0) ? current->next : current->prev;
    }

    return static_cast<const Entry>(current->entry);
}

/**
 * Removes the given element from the list.
 * It assumes, that the given Element is part
 * of the list (it DOES NOT check it).
 */
template<class Entry>
void DLList<Entry>::remove(Element<Entry>* elem) {

    if (elements > 1) {
        elem->prev->next = elem->next;
        elem->next->prev = elem->prev;

        if (elem == first) first = elem->next;
        if (elem == last)  last  = elem->prev;

    } else {
        first = NULL;
        last  = NULL;
    }

    delete elem;
    elements--;
}

template<class Entry>
Entry DLList<Entry>::pop_back() {

    Entry e = last->entry;
    remove(last);
    return e;
}

template<class Entry>
Entry DLList<Entry>::pop_front() {

    Entry e = first->entry;
    remove(first);
    return e;
}

template<class Entry>
DLListIterator<Entry>* DLList<Entry>::iterator() {
    return new DLListIterator<Entry>(this);
}

template<class Entry>
DLListRIterator<Entry>* DLList<Entry>::riterator() {
    return new DLListRIterator<Entry>(this);
}

/**
 * Quicksort the elements by their entries '<' relation.
 * If the elements are pointers, compare their targets.
 */
template<class Entry>
void DLList<Entry>::sort() {

    if (elements > 1)
        this->sort(first, last);
}

template<class Entry>
void DLList<Entry>::sort(Element<Entry>* l, Element<Entry>* r) {

    Element<Entry>* splitter = r;
    Element<Entry>* i = l->prev;
    Element<Entry>* j = r;

    for (;;) {
        do { i = i->next; } while (i->entry < splitter->entry);
        do { j = j->prev; } while (splitter->entry < j->entry);
        if ((i == j) || (i == j->next)) break;
        exchange(i, j);
    }
    exchange(i, r);

    if ((l != j) && (l != j->next)) sort(l, i->prev);
    if ((r != i) && (r != i->prev)) sort(i->next, r);
}

/**
 * Print the list to stdout.
 */
template<class Entry>
void DLList<Entry>::print() {

    Element<Entry>* cur = first;

    std::cout << "************** print: standard" << std::endl;

    if (empty())
        std::cout << "List empty." << std::endl;
    else do {
        std::cout << "Element: " << cur->entry << " next: " << cur->next << " last: " << cur->prev << std::endl;
        cur = cur->next;
    } while ((cur != NULL) && (cur != last->next));
}

/**
 * Push all elements of 'list' to the end of
 * this list. 'list' will be empty after this
 * operation.
 */
template<class Entry>
void DLList<Entry>::push_back(DLList<Entry>* list) {

    /*while(!list->empty()) {
        push_back(list->pop_front());
    }*/

    if (!list->empty()) {
        if (!empty()) {
            last->next = list->first;
            list->first->prev = last;
            first->prev = list->last;
            list->last->next = first;

            last = list->last;
        } else {
            first = list->first;
            last = list->last;
        }
        elements += list->elements;
        list->first = NULL;
        list->last = NULL;
        list->elements = 0;
    }
}
/******* Pointer implementations *********************************************************/

template<class Entry>
DLList<Entry*>::DLList() {

    first = NULL;
    last = NULL;
    elements = 0;
}

template<class Entry>
DLList<Entry*>::~DLList() {

    this->clear();
}

template<class Entry>
void DLList<Entry*>::clear() {

    while (!empty())
        remove(last);

    first = NULL;
    last  = NULL;
}

/**
 * A helper function to hold shared code of both
 * push_back and push_front.
 * It adds the new entry between back and front.
 */
template<class Entry>
void DLList<Entry*>::ring_insert(Entry* e) {

    Element<Entry*>* elem = new Element<Entry*>(e);

    if (empty()) {
        first = elem;
        last  = elem;
        elem->next = elem;
        elem->prev = elem;
    } else {
        last->next = elem;
        elem->prev = last;
        first->prev = elem;
        elem->next = first;
    }

    elements++;
}

template<class Entry>
void DLList<Entry*>::push_back(Entry* e) {

    ring_insert(e);
    last = last->next;
}

template<class Entry>
void DLList<Entry*>::push_front(Entry* e) {

    ring_insert(e);
    first = first->prev;
}

template<class Entry>
Entry* DLList<Entry*>::at(unsigned long pos) const{

    if ((pos > elements-1) || (pos < 0) || (empty()))
        return NULL;

    int dec = 1;
    Element<Entry>* current = first;

    if (pos > elements / 2) {
        dec = -1;
        current = last;
        pos = pos - elements;
    }

    while (pos != 0) {
        pos = pos - dec;
        current = (dec > 0) ? current->next : current->prev;
    }

    return static_cast<const Entry>(current->entry);
}

/**
 * Removes the given element from the list.
 * It assumes, that the given Element is part
 * of the list (it DOES NOT check it).
 */
template<class Entry>
void DLList<Entry*>::remove(Element<Entry*>* elem) {

    if (elements > 1) {
        elem->prev->next = elem->next;
        elem->next->prev = elem->prev;

        if (elem == first) first = elem->next;
        if (elem == last)  last  = elem->prev;

    } else {
        first = NULL;
        last  = NULL;
    }

    delete elem;
    elements--;
}

template<class Entry>
Entry* DLList<Entry*>::pop_back() {

    Entry* e = last->entry;
    remove(last);
    return e;
}

template<class Entry>
Entry* DLList<Entry*>::pop_front() {

    Entry* e = first->entry;
    remove(first);
    return e;
}

template<class Entry>
DLListIterator<Entry*>* DLList<Entry*>::iterator() {
    return new DLListIterator<Entry*>(this);
}

template<class Entry>
DLListRIterator<Entry*>* DLList<Entry*>::riterator() {
    return new DLListRIterator<Entry*>(this);
}

/**
 * Quicksort the elements by their entries '<' relation.
 * If the elements are pointers, compare their targets.
 */
template<class Entry>
void DLList<Entry*>::sort() {

    if (elements > 1)
        this->sort(first, last);
}

template<class Entry>
void DLList<Entry*>::sort(Element<Entry*>* l, Element<Entry*>* r) {

    Element<Entry*>* splitter = r;
    Element<Entry*>* i = l->prev;
    Element<Entry*>* j = r;

    for (;;) {
        do { i = i->next; } while (*(i->entry) < *(splitter->entry));
        do { j = j->prev; } while (*(splitter->entry) < *(j->entry));
        if ((i == j) || (i == j->next)) break;
        exchange(i, j);
    }
    exchange(i, r);

    if ((l != j) && (l != j->next)) sort(l, i->prev);
    if ((r != i) && (r != i->prev)) sort(i->next, r);
}

template<class Entry>
void DLList<Entry*>::print() {

    Element<Entry*>* cur = first;

    std::cout << "************** print: pointers" << std::endl;

    if (empty())
        std::cout << "List empty." << std::endl;
    else do {
        std::cout << "Element: " << *(cur->entry) << " next: " << cur->next << " last: " << cur->prev << std::endl;
        cur = cur->next;
    } while ((cur != NULL) && (cur != last->next));
}

template<class Entry>
void DLList<Entry*>::push_back(DLList<Entry*>* list) {

    if (!list->empty()) {
        if (!empty()) {
            last->next = list->first;
            list->first->prev = last;
            first->prev = list->last;
            list->last->next = first;

            last = list->last;
        } else {
            first = list->first;
            last = list->last;
        }
        elements += list->elements;
        list->first = NULL;
        list->last = NULL;
        list->elements = 0;
    }
}
